# Import packages in reverse libOrder as defined in pyqtgraph/Qt/__init__.py
from PySide2.QtCore import *
from PyQt5.QtCore import *
from PySide6.QtCore import *
from PyQt6.QtCore import *
Signal = pyqtSignal
Slot = pyqtSlot
