GL_VENDOR = 0x1F00
GL_RENDERER = 0x1F01
GL_VERSION = 0x1F02
GL_SHADING_LANGUAGE_VERSION = 0x8B8C

GL_COLOR_BUFFER_BIT   = 0x00004000
GL_STENCIL_BUFFER_BIT = 0x00000400
GL_DEPTH_BUFFER_BIT   = 0x00000100

GL_LINES = 0x0001
GL_LINE_LOOP = 0x0002
GL_LINE_STRIP = 0x0003
GL_TRIANGLES = 0x0004
GL_TRIANGLE_STRIP = 0x0005
GL_NEVER = 0x0200
GL_EQUAL = 0x0202
GL_ALWAYS = 0x0207
GL_SRC_ALPHA = 0x0302
GL_ONE_MINUS_SRC_ALPHA = 0x0303
GL_LINE_SMOOTH = 0x0B20
GL_STENCIL_TEST = 0x0B90
GL_SCISSOR_TEST = 0x0C11
GL_BLEND = 0x0BE2
GL_LINE_SMOOTH_HINT = 0x0C52
GL_NICEST = 0x1102
GL_UNSIGNED_INT = 0x1405
GL_FLOAT = 0x1406
GL_KEEP = 0x1E00
GL_REPLACE = 0x1E01
