# Import packages in reverse libOrder as defined in pyqtgraph/Qt/__init__.py
from PySide2.QtWidgets import *
from PyQt5.QtWidgets import *
from PySide6.QtWidgets import *
from PyQt6.QtWidgets import *
